package ru.dbotthepony.mc.otm.block.entity.matter

import com.google.common.collect.Streams
import javax.annotation.ParametersAreNonnullByDefault
import net.minecraft.core.BlockPos
import net.minecraft.world.level.block.state.BlockState
import ru.dbotthepony.mc.otm.container.MatteryContainer
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.block.matter.PatternStorageBlock
import net.minecraft.nbt.CompoundTag
import net.minecraftforge.common.util.LazyOptional
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import ru.dbotthepony.mc.otm.menu.PatternStorageMenu
import net.minecraft.MethodsReturnNonnullByDefault
import net.minecraft.core.Direction
import net.minecraft.network.chat.Component
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.Container
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Block
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ForgeCapabilities
import ru.dbotthepony.mc.otm.block.IDroppableContainer
import ru.dbotthepony.mc.otm.block.entity.MatteryBlockEntity
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.capability.matter.*
import ru.dbotthepony.mc.otm.core.iterator
import ru.dbotthepony.mc.otm.graph.Graph6Node
import ru.dbotthepony.mc.otm.graph.matter.IMatterGraphNode
import ru.dbotthepony.mc.otm.graph.matter.MatterNetworkGraph
import ru.dbotthepony.mc.otm.core.map
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import ru.dbotthepony.mc.otm.core.set
import java.util.ArrayList
import java.util.stream.Stream

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
class PatternStorageBlockEntity(p_155229_: BlockPos, p_155230_: BlockState) :
	MatteryBlockEntity(MBlockEntities.PATTERN_STORAGE, p_155229_, p_155230_), IMatterGraphNode, IPatternStorage, IDroppableContainer {

	override val matterNode = Graph6Node<IMatterGraphNode>(this)
	private val resolverPatterns = LazyOptional.of { this }
	private val resolverNode = LazyOptional.of { this }

	val patternContainer: MatteryContainer = object : MatteryContainer(this::setChanged, 8) {
		override fun setChanged(slot: Int, new: ItemStack, old: ItemStack) {
			val grid = matterNode.graph as MatterNetworkGraph?

			if (grid != null && !ItemStack.isSameItemSameTags(new, old)) {
				if (!old.isEmpty) {
					old.getCapability(MatteryCapability.PATTERN).ifPresent { cap: IPatternStorage ->
						cap.patterns.forEach { grid.onPatternRemoved(it) }
					}
				}

				if (!new.isEmpty) {
					new.getCapability(MatteryCapability.PATTERN).ifPresent { cap: IPatternStorage ->
						cap.patterns.forEach { grid.onPatternAdded(it) }
					}
				}

				updateBlockstate()
			}

			super.setChanged(slot, new, old)
		}

		override fun getMaxStackSize(): Int {
			return 1
		}
	}

	private fun updateBlockstate() {
		val level = level ?: return

		var state = blockState

		for (i in 0..7) {
			state = state.setValue(
				PatternStorageBlock.PATTERN_STORAGE_DISKS_PROPS[i],
				this.patternContainer.getItem(i).getCapability(MatteryCapability.PATTERN).isPresent
			)
		}

		if (state !== blockState) {
			level.setBlock(blockPos, state, Block.UPDATE_CLIENTS)
		}
	}

	private val resolverItem =
		patternContainer.handler { slot: Int, stack: ItemStack -> stack.getCapability(MatteryCapability.PATTERN).isPresent }

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt[INVENTORY_KEY] = patternContainer.serializeNBT()
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)
		nbt.map(INVENTORY_KEY, patternContainer::deserializeNBT)
	}

	override val droppableContainer: Container
		get() = patternContainer

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)

		if (p_155231_ is ServerLevel)
			MatterNetworkGraph.discoverFull(this, matterNode)
	}

	override fun getPatternHandler(): IPatternStorage {
		return this
	}

	private var valid = true

	override fun invalidateCaps() {
		super.invalidateCaps()
		valid = false
		resolverItem.invalidate()
		matterNode.destroy(::MatterNetworkGraph)
	}

	override fun reviveCaps() {
		super.reviveCaps()
		valid = true
		resolverItem.revive()
	}

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (valid) {
			if (cap == MatteryCapability.PATTERN) return resolverPatterns.cast()
			if (cap == MatteryCapability.MATTER_NODE) return resolverNode.cast()
			if (cap == ForgeCapabilities.ITEM_HANDLER) return resolverItem.get().cast()
		}

		return super.getCapability(cap, side)
	}

	override val defaultDisplayName: Component
		get() = MACHINE_NAME

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return PatternStorageMenu(containerID, inventory, this)
	}

	override val patterns: Stream<out IPatternState> get() {
		val streams = ArrayList<Stream<out IPatternState>>()

		for (provider in this.patternContainer.iterator(MatteryCapability.PATTERN)) {
			streams.add(provider.second.patterns)
		}

		return Streams.concat(*streams.toTypedArray())
	}

	override val patternCapacity: Int get() {
		var stored = 0L

		for ((_, pattern) in this.patternContainer.iterator(MatteryCapability.PATTERN))
			stored += pattern.patternCapacity.toLong()

		return if (stored > Int.MAX_VALUE) Int.MAX_VALUE else stored.toInt()
	}

	override val storedPatterns: Int get() {
		var stored = 0L

		for ((_, pattern) in this.patternContainer.iterator(MatteryCapability.PATTERN))
			stored += pattern.storedPatterns.toLong()

		return if (stored > Int.MAX_VALUE) Int.MAX_VALUE else stored.toInt()
	}

	override fun setRemoved() {
		super.setRemoved()
		matterNode.destroy(::MatterNetworkGraph)
	}

	override fun insertPattern(pattern: IPatternState, onlyUpdate: Boolean, simulate: Boolean): PatternInsertStatus {
		for (pair in this.patternContainer.iterator(MatteryCapability.PATTERN)) {
			val status = pair.second.insertPattern(pattern, onlyUpdate, simulate)

			if (!status.isFailed) {
				if (!simulate) {
					setChanged()

					val graph = matterNode.graph as MatterNetworkGraph?

					if (graph != null) {
						if (status.isInserted) {
							graph.onPatternAdded(status.newState!!)
						} else {
							graph.onPatternUpdated(status.newState!!, status.oldState!!)
						}
					}
				}

				return status
			}
		}

		return PatternInsertFailure
	}

	companion object {
		private val MACHINE_NAME = TranslatableComponent("block.overdrive_that_matters.pattern_storage")
	}
}
