package ru.dbotthepony.mc.otm.core

import net.minecraft.nbt.ByteArrayTag
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.StringTag
import net.minecraft.nbt.Tag
import net.minecraft.network.FriendlyByteBuf
import net.minecraftforge.common.ForgeConfigSpec
import ru.dbotthepony.mc.otm.ObservedConfigValue
import java.io.InputStream
import java.io.OutputStream
import java.math.BigDecimal
import java.math.BigInteger
import java.math.MathContext
import kotlin.math.absoluteValue

//private fun isZero(value: BigInteger) = value == BigInteger.ZERO
private fun isZero(value: BigInteger) = value.signum() == 0

private fun unsignedInt(value: Byte): Int {
	return value.toInt() and 0xFF
}

private fun usignedLong(value: Byte): Long {
	return value.toLong() and 0xFFL
}

private fun longToBytesBE(value: Long): ByteArray {
	return byteArrayOf(
		(value ushr 56).toByte(),
		(value ushr 48).toByte(),
		(value ushr 40).toByte(),
		(value ushr 32).toByte(),
		(value ushr 24).toByte(),
		(value ushr 16).toByte(),
		(value ushr 8).toByte(),
		(value).toByte(),
	)
}

private fun bytesToLongBE(value: ByteArray, from: Int = 0): Long {
	return (
		(usignedLong(value[from + 7])) or
		(usignedLong(value[from + 6]) shl 8) or
		(usignedLong(value[from + 5]) shl 16) or
		(usignedLong(value[from + 4]) shl 24) or
		(usignedLong(value[from + 3]) shl 32) or
		(usignedLong(value[from + 2]) shl 40) or
		(usignedLong(value[from + 1]) shl 48) or
		(usignedLong(value[from]) shl 56)
	)
}

private fun bytesToLongBE(
	value7: Byte,
	value6: Byte,
	value5: Byte,
	value4: Byte,
	value3: Byte,
	value2: Byte,
	value1: Byte,
	value0: Byte,
): Long {
	return (
		(usignedLong(value0)) or
		(usignedLong(value1) shl 8) or
		(usignedLong(value2) shl 16) or
		(usignedLong(value3) shl 24) or
		(usignedLong(value4) shl 32) or
		(usignedLong(value5) shl 40) or
		(usignedLong(value6) shl 48) or
		(usignedLong(value7) shl 56)
	)
}

/**
 * Constant value which represent edge of meaningful bits.
 *
 * Equals to 0.000000000001
 */
const val EPSILON = 0.000000000001

private val zeroes = Array(16) { "0".repeat(it).intern() }
private val nums = Array(10) { it.toString()[0] }

private fun decimals(input: Double, places: Int): String {
	if (weakEqualDoubles(input, 0.0)) {
		return zeroes.elementAtOrElse(places) { "0".repeat(places) }
	}

	val buffer = CharArray(places)

	@Suppress("NAME_SHADOWING")
	var input = input.absoluteValue

	for (i in 0 until places) {
		input *= 10.0
		buffer[i] = nums[(input % 10.0).toInt()]
	}

	return String(buffer)
}

fun weakEqualDoubles(a: Double, b: Double): Boolean {
	if (a == b)
		return true

	return (a - b).absoluteValue <= EPSILON
}

fun weakCompareDoubles(a: Double, b: Double): Int {
	if (weakEqualDoubles(a, b))
		return 0

	if (a > b)
		return 1

	return -1
}

fun weakLessThan(a: Double, b: Double) = weakCompareDoubles(a, b) < 0
fun weakGreaterThan(a: Double, b: Double) = weakCompareDoubles(a, b) > 0
fun weakLessOrEqual(a: Double, b: Double) = weakCompareDoubles(a, b) <= 0
fun weakGreaterOrEqual(a: Double, b: Double) = weakCompareDoubles(a, b) >= 0

private val PERCENTAGE_CONTEXT = MathContext(6)

private const val MEANINGFUL_BITS_LONG = 1000000000000000000L
private const val MEANINGFUL_BITS_DOUBLE = MEANINGFUL_BITS_LONG.toDouble()
private val MEANINGFUL_BITS_BI = BigInteger.valueOf(MEANINGFUL_BITS_LONG)

private val BI_MINUS_ONE = -BigInteger.ONE

fun ImpreciseFraction(value: Byte) = ImpreciseFraction.valueOf(value)
fun ImpreciseFraction(value: Short) = ImpreciseFraction.valueOf(value)
fun ImpreciseFraction(value: Int) = ImpreciseFraction.valueOf(value)
fun ImpreciseFraction(value: Long) = ImpreciseFraction.valueOf(value)

/**
 * Imprecise Fraction class for dealing with huge numbers that need fractional part to some extent.
 *
 * In essence, this class is pretty much like [BigDecimal], except decimal part of this number is
 * not guaranteed to be precise (stored as [Double]). The reason behind creation of this class, however,
 * is to allow infinite precision of [whole] part, while leaving [decimal] part to be rounded in inexact operations.
 *
 * This class is value based, however, [equals] and [compareTo] are not doing *exact* comparison. Whole part is always compared
 * exactly, but decimal part is considered equal if their difference is less than [EPSILON].
 *
 *
 */
@Suppress("unused")
class ImpreciseFraction @JvmOverloads constructor(whole: BigInteger, decimal: Double = 0.0) : Number(), Comparable<ImpreciseFraction>, java.io.Serializable {
	constructor(whole: Byte, decimal: Double) : this(BigInteger.valueOf(whole.toLong()), decimal)
	constructor(whole: Short, decimal: Double) : this(BigInteger.valueOf(whole.toLong()), decimal)
	constructor(whole: Int, decimal: Double) : this(BigInteger.valueOf(whole.toLong()), decimal)
	constructor(whole: Long, decimal: Double) : this(BigInteger.valueOf(whole), decimal)
	constructor(value: BigDecimal) : this(value.toBigInteger(), value.remainder(BigDecimal.ONE).toDouble()) // TODO
	constructor(value: Float) : this(BigInteger.valueOf((value - value % 1.0).toLong()), value % 1.0)
	constructor(value: Double) : this(BigInteger.valueOf((value - value % 1.0).toLong()), value % 1.0)
	constructor(value: String) : this(BigDecimal(value))

	/**
	 * Inexact fractional part of this fraction
	 */
	val decimal: Double

	/**
	 * Exact whole part of this fraction
	 */
	val whole: BigInteger

	init {
		@Suppress("name_shadowing")
		var decimal = decimal
		@Suppress("name_shadowing")
		var whole = whole

		if (decimal <= -1f || decimal >= 1f) {
			val frac = decimal % 1f
			whole += BigInteger.valueOf((decimal - frac).toLong())
			decimal = frac
		}

		// дробная часть равна или очень близка к нулю
		if (weakEqualDoubles(decimal, 0.0)) {
			if (!isZero(whole)) {
				this.decimal = 0.0
				this.whole = whole
			} else { // сохраняет минус ноль
				this.decimal = decimal
				this.whole = whole
			}
		// целая часть и дробная часть имеют одинаковые знаки
		} else if (decimal > 0.0 && whole.signum() >= 0 || decimal < 0.0 && whole.signum() <= 0) {
			this.decimal = decimal
			this.whole = whole
		} else {
			// целая часть и дробная часть имеют разные знаки

			if (decimal > 0.0) {
				this.decimal = decimal - 1f
				this.whole = whole + BigInteger.ONE
			//} else if (double < 0.0 && whole.signum() > 0) {
			} else {
				this.decimal = decimal + 1f
				this.whole = whole - BigInteger.ONE
			}

			//throw InternalError("This should never happen, but it happened. Input was: $whole $double")
		}
	}

	val isNaN get() = decimal.isNaN()

	operator fun plus(other: ImpreciseFraction) = ImpreciseFraction(whole + other.whole, decimal + other.decimal)
	operator fun minus(other: ImpreciseFraction) = ImpreciseFraction(whole - other.whole, decimal - other.decimal)

	operator fun times(other: ImpreciseFraction): ImpreciseFraction {
		if (other == ONE)
			return this
		else if (other == MINUS_ONE)
			return -this

		val a = whole
		val c = other.whole

		val bZero = weakEqualDoubles(decimal, 0.0)
		val dZero = weakEqualDoubles(other.decimal, 0.0)

		if (bZero && dZero) {
			if (isZero(a) || isZero(c))
				return ZERO

			return ImpreciseFraction(a * c)
		} else if (isZero(a) && isZero(c)) {
			if (bZero || dZero)
				return ZERO

			return ImpreciseFraction(0, decimal * other.decimal)
		} else if (bZero) {
			val d = other.decimal * MEANINGFUL_BITS_DOUBLE
			val dL = d.toLong()
			val adInflated = a * BigInteger.valueOf(dL)
			val ad = adInflated.divideAndRemainder(MEANINGFUL_BITS_BI)

			if (isZero(c)) {
				return ImpreciseFraction(
					ad[0],
					ad[1].toDouble() / MEANINGFUL_BITS_DOUBLE
				)
			}

			return ImpreciseFraction(
				a * c + ad[0],
				ad[1].toDouble() / MEANINGFUL_BITS_DOUBLE
			)
		} else if (dZero) {
			val b = decimal * MEANINGFUL_BITS_DOUBLE
			val bL = b.toLong()
			val bcInflated = c * BigInteger.valueOf(bL)
			val bc = bcInflated.divideAndRemainder(MEANINGFUL_BITS_BI)

			if (isZero(a)) {
				return ImpreciseFraction(
					bc[0],
					bc[1].toDouble() / MEANINGFUL_BITS_DOUBLE
				)
			}

			return ImpreciseFraction(
				a * c + bc[0],
				bc[1].toDouble() / MEANINGFUL_BITS_DOUBLE
			)
		}

		val b = decimal * MEANINGFUL_BITS_DOUBLE
		val d = other.decimal * MEANINGFUL_BITS_DOUBLE

		val bL = b.toLong()
		val dL = d.toLong()

		val bcInflated = c * BigInteger.valueOf(bL)
		val adInflated = a * BigInteger.valueOf(dL)

		val bc = bcInflated.divideAndRemainder(MEANINGFUL_BITS_BI)
		val ad = adInflated.divideAndRemainder(MEANINGFUL_BITS_BI)

		return ImpreciseFraction(
			a * c + bc[0] + ad[0],
			decimal * other.decimal + bc[1].toDouble() / MEANINGFUL_BITS_DOUBLE + ad[1].toDouble() / MEANINGFUL_BITS_DOUBLE
		)
	}

	operator fun div(other: ImpreciseFraction): ImpreciseFraction {
		if (other == ONE)
			return this
		else if (other == MINUS_ONE)
			return -this

		if (isZero && other.isZero)
			return NaN

		if (!isZero && other.isZero)
			throw ArithmeticException("Divide by zero")

		//if (isZero(whole) && isZero(other.whole))
		//	return ImpreciseFraction(BigInteger.ZERO, decimal / other.decimal)

		val a = toBigDecmial()
		val b = other.toBigDecmial()
		val div = a.divideAndRemainder(b, MathContext(a.precision() + b.precision()))

		val bD = b.toDouble()

		if (bD.isInfinite() || bD == 0.0) {
			return ImpreciseFraction(div[0].toBigInteger())
		}

		return ImpreciseFraction(div[0].toBigInteger(), div[1].toDouble() / bD)
	}

	operator fun plus(other: Float): ImpreciseFraction {
		if (other == 0f) {
			return this
		}

		return plus(ImpreciseFraction(other))
	}

	operator fun minus(other: Float): ImpreciseFraction {
		if (other == 0f) {
			return this
		}

		return minus(ImpreciseFraction(other))
	}

	operator fun times(other: Float): ImpreciseFraction {
		if (other == 1f) {
			return this
		} else if (other == 0f) {
			return ZERO
		} else if (other == -1f) {
			return -this
		}

		return times(ImpreciseFraction(other))
	}

	operator fun div(other: Float): ImpreciseFraction {
		if (other == 0f) {
			if (isZero)
				return NaN
			else
				throw ArithmeticException("Divide by zero")
		} else if (other == 1f) {
			return this
		} else if (other == -1f) {
			return -this
		}

		return div(ImpreciseFraction(other))
	}

	operator fun plus(other: Double): ImpreciseFraction {
		if (other == 0.0) {
			return this
		}

		return plus(ImpreciseFraction(other))
	}

	operator fun minus(other: Double): ImpreciseFraction {
		if (other == 0.0) {
			return this
		}

		return minus(ImpreciseFraction(other))
	}

	operator fun times(other: Double): ImpreciseFraction {
		if (other == 1.0) {
			return this
		} else if (other == 0.0) {
			return ZERO
		} else if (other == -1.0) {
			return -this
		}

		return times(ImpreciseFraction(other))
	}

	operator fun div(other: Double): ImpreciseFraction {
		if (other == 0.0) {
			if (isZero)
				return NaN
			else
				throw ArithmeticException("Divide by zero")
		} else if (other == 1.0) {
			return this
		} else if (other == -1.0) {
			return -this
		}

		return div(ImpreciseFraction(other))
	}

	operator fun plus(other: Int): ImpreciseFraction {
		if (other == 0) {
			return this
		}

		return plus(valueOf(other))
	}

	operator fun minus(other: Int): ImpreciseFraction {
		if (other == 0) {
			return this
		}

		return minus(valueOf(other))
	}

	operator fun times(other: Int): ImpreciseFraction {
		if (other == 1) {
			return this
		} else if (other == 0) {
			return ZERO
		} else if (other == -1) {
			return -this
		}

		return times(valueOf(other))
	}

	operator fun div(other: Int): ImpreciseFraction {
		if (other == 0) {
			if (isZero)
				return NaN
			else
				throw ArithmeticException("Divide by zero")
		} else if (other == 1) {
			return this
		} else if (other == -1) {
			return -this
		}

		return div(valueOf(other))
	}

	operator fun plus(other: Long): ImpreciseFraction {
		if (other == 0L) {
			return this
		}

		return plus(valueOf(other))
	}

	operator fun minus(other: Long): ImpreciseFraction {
		if (other == 0L) {
			return this
		}

		return minus(valueOf(other))
	}

	operator fun times(other: Long): ImpreciseFraction {
		if (other == 1L) {
			return this
		} else if (other == 0L) {
			return ZERO
		} else if (other == -1L) {
			return -this
		}

		return times(valueOf(other))
	}

	operator fun div(other: Long): ImpreciseFraction {
		if (other == 0L) {
			if (isZero)
				return NaN
			else
				throw ArithmeticException("Divide by zero")
		} else if (other == 1L) {
			return this
		} else if (other == -1L) {
			return -this
		}

		return div(valueOf(other))
	}

	operator fun plus(other: BigInteger): ImpreciseFraction {
		if (other == BigInteger.ZERO) {
			return this
		}

		return plus(ImpreciseFraction(other))
	}

	operator fun minus(other: BigInteger): ImpreciseFraction {
		if (other == BigInteger.ZERO) {
			return this
		}

		return minus(ImpreciseFraction(other))
	}

	operator fun times(other: BigInteger): ImpreciseFraction {
		if (other == BigInteger.ONE) {
			return this
		} else if (other == BigInteger.ZERO) {
			return ZERO
		} else if (other == BI_MINUS_ONE) {
			return -this
		}

		return times(ImpreciseFraction(other))
	}

	operator fun div(other: BigInteger): ImpreciseFraction {
		if (other == BigInteger.ZERO) {
			return NaN
		} else if (other == BigInteger.ONE) {
			return this
		} else if (other == BI_MINUS_ONE) {
			return -this
		}

		return div(ImpreciseFraction(other))
	}

	operator fun unaryMinus(): ImpreciseFraction {
		return ImpreciseFraction(-whole, -decimal)
	}

	override fun equals(other: Any?): Boolean {
		if (isNaN)
			return false

		if (other is ImpreciseFraction) {
			return other.whole == whole && weakEqualDoubles(decimal, other.decimal)
		}

		return super.equals(other)
	}

	fun equalsStrict(other: Any?): Boolean {
		if (isNaN)
			return other is ImpreciseFraction && other.isNaN // ибо hashCode() так требует

		if (other is ImpreciseFraction) {
			return other.whole == whole && other.decimal == decimal
		}

		return super.equals(other)
	}

	override fun hashCode(): Int {
		if (isNaN)
			return Double.NaN.hashCode()

		return 31 * (decimal - decimal % EPSILON).hashCode() + whole.hashCode()
	}

	fun floor(): ImpreciseFraction {
		if (decimal == 0.0)
			return this

		return ImpreciseFraction(whole)
	}

	fun toString(decimals: Int): String {
		if (isNaN)
			return "NaN"

		if (decimals == 0) {
			return whole.toString()
		} else if (decimals > 0) {
			return "$whole.${decimals(decimal, decimals)}"
		}

		return when (signum()) {
			1, -1 -> "$whole.${decimals(decimal, 11)}"
			0 -> "0.0"
			else -> throw IllegalArgumentException("invalid signum")
		}
	}

	override fun toString() = toString(-1)

	fun toBigDecmial(): BigDecimal {
		return BigDecimal(whole) + BigDecimal(decimal)
	}

	fun signum(): Int {
		val sign = whole.signum()

		if (sign != 0) {
			return sign
		}

		// exactly zero
		if (decimal == 0.0)
			return if (1.0 / decimal > 0) 1 else -1

		// inexactly zero
		if (weakEqualDoubles(decimal, 0.0))
			return 0

		return if (decimal > 0.0) 1 else -1
	}

	override fun compareTo(other: ImpreciseFraction): Int {
		if (isNaN)
			return 1
		else if (other.isNaN)
			return -1

		val a = signum()
		val b = other.signum()

		if (a < b)
			return -1
		else if (a > b)
			return 1

		if (other.whole == whole) {
			return weakCompareDoubles(decimal, other.decimal)
		}

		return whole.compareTo(other.whole)
	}

	fun compareToStrict(other: ImpreciseFraction): Int {
		if (isNaN)
			return 1
		else if (other.isNaN)
			return -1

		val a = signum()
		val b = other.signum()

		if (a < b)
			return -1
		else if (a > b)
			return 1

		if (other.whole == whole) {
			if (other.decimal == decimal) {
				return 0
			}

			return if (other.decimal < decimal) 1 else -1
		}

		return whole.compareTo(other.whole)
	}

	fun toByteArray(): ByteArray {
		val whole = whole.toByteArray()
		return byteArrayOf(whole.size.toByte(), *whole, *longToBytesBE(decimal.toBits()))
	}

	fun serializeNBT(): Tag {
		return ByteArrayTag(toByteArray())
	}

	val isZero get() = weakEqualDoubles(decimal, 0.0) && isZero(whole)
	val isPositive get() = this > ZERO
	val isNegative get() = this < ZERO

	fun moreThanZero(): ImpreciseFraction {
		if (signum() >= 0)
			return this

		return ZERO
	}

	fun lessOrZero(): ImpreciseFraction {
		if (signum() <= 0)
			return this

		return ZERO
	}

	override fun toInt(): Int {
		if (whole > BI_INT_MAX) {
			return Int.MAX_VALUE
		} else if (whole < BI_INT_MIN) {
			return Int.MIN_VALUE
		}

		return whole.toInt()
	}

	fun write(buff: FriendlyByteBuf) {
		val whole = whole.toByteArray()
		buff.writeByteArray(whole)
		buff.writeDouble(decimal)
	}

	override fun toFloat(): Float {
		return whole.toFloat() + decimal.toFloat()
	}

	override fun toDouble(): Double {
		return whole.toDouble() + decimal
	}

	override fun toByte(): Byte {
		return toInt().toByte()
	}

	override fun toChar(): Char {
		return toInt().toChar()
	}

	override fun toLong(): Long {
		if (whole > BI_LONG_MAX) {
			return Long.MAX_VALUE
		} else if (whole < BI_LONG_MIN) {
			return Long.MIN_VALUE
		}

		return whole.toLong()
	}

	override fun toShort(): Short {
		return toInt().toShort()
	}

	fun toFraction(): Fraction {
		if (isZero) {
			return Fraction.ZERO
		}

		if (weakEqualDoubles(decimal, 0.0)) {
			return Fraction(whole)
		}

		return Fraction(toBigDecmial())
	}

	fun percentage(divisor: ImpreciseFraction, zeroing: Boolean = true): Float {
		if ((isZero || divisor.isZero) && zeroing) return 0f
		if (isNaN || divisor.isNaN || divisor.isZero) return Float.NaN

		if (this >= divisor)
			return 1f

		if (this <= FLOAT_MAX_VALUE && divisor <= FLOAT_MAX_VALUE)
			return (toDouble() / divisor.toDouble()).toFloat()

		return toBigDecmial().divide(divisor.toBigDecmial(), PERCENTAGE_CONTEXT).toFloat()
	}

	operator fun rem(divisor: ImpreciseFraction): ImpreciseFraction {
		TODO("Not yet implemented")
	}

	companion object {
		@JvmStatic
		val serialVersionUID: Long = 287354739494574838L

		@JvmField val MINUS_ZERO = ImpreciseFraction(0, -0.0)
		@JvmField val ZERO = ImpreciseFraction(BigInteger.ZERO)
		@JvmField val ONE = ImpreciseFraction(BigInteger.ONE)
		@JvmField val MINUS_ONE = ImpreciseFraction(-BigInteger.ONE)
		@JvmField val TWO = ImpreciseFraction(BigInteger.TWO)
		@JvmField val TEN = ImpreciseFraction(BigInteger.TEN)

		@JvmField val INT_MAX_VALUE = ImpreciseFraction(BI_INT_MAX)
		@JvmField val INT_MIN_VALUE = ImpreciseFraction(BI_INT_MIN)
		@JvmField val LONG_MAX_VALUE = ImpreciseFraction(BI_LONG_MAX)
		@JvmField val LONG_MIN_VALUE = ImpreciseFraction(BI_LONG_MIN)

		@JvmField val FLOAT_MAX_VALUE = ImpreciseFraction(BI_FLOAT_MAX)
		@JvmField val FLOAT_MIN_VALUE = ImpreciseFraction(BI_FLOAT_MIN)
		@JvmField val DOUBLE_MAX_VALUE = ImpreciseFraction(BI_DOUBLE_MAX)
		@JvmField val DOUBLE_MIN_VALUE = ImpreciseFraction(BI_DOUBLE_MIN)

		@JvmField val ONE_TENTH = ImpreciseFraction("0.1")

		private val cache = Array(2048) { ImpreciseFraction(it - 1024, 0.0) }

		/**
		 * Returns pooled value if present, otherwise constructs new object
		 */
		fun valueOf(value: Int): ImpreciseFraction {
			if (value in -1024 .. 1023) {
				return cache[value + 1024]
			}

			return ImpreciseFraction(value, 0.0)
		}

		/**
		 * Returns pooled value if present, otherwise constructs new object
		 */
		fun valueOf(value: Long): ImpreciseFraction {
			if (value in -1024L .. 1023L) {
				return cache[value.toInt() + 1024]
			}

			return ImpreciseFraction(value, 0.0)
		}

		/**
		 * Returns pooled value if present, otherwise constructs new object
		 */
		fun valueOf(value: Short) = valueOf(value.toInt())

		/**
		 * Returns pooled value if present, otherwise constructs new object
		 */
		fun valueOf(value: Byte) = valueOf(value.toInt())

		@JvmField
		val NaN = ImpreciseFraction(0, Double.NaN)

		@JvmStatic
		fun fromByteArray(input: ByteArray): ImpreciseFraction {
			val size = unsignedInt(input[0])

			if (size > 0) {
				val slice = input.copyOfRange(1, 1 + size)
				val bits = bytesToLongBE(input, 1 + size)
				return ImpreciseFraction(BigInteger(slice), Double.fromBits(bits))
			} else {
				return ImpreciseFraction(BigInteger.ZERO, Double.fromBits(bytesToLongBE(input, 1)))
			}
		}

		@JvmStatic
		fun deserializeNBT(input: Tag?): ImpreciseFraction {
			if (input is ByteArrayTag) {
				return fromByteArray(input.asByteArray)
			} else if (input is StringTag) {
				return ImpreciseFraction(input.asString)
			}

			return ZERO
		}

		@JvmStatic
		fun deserializeNBT(input: ByteArrayTag): ImpreciseFraction {
			return fromByteArray(input.asByteArray)
		}

		@JvmStatic
		fun read(buff: FriendlyByteBuf): ImpreciseFraction {
			return ImpreciseFraction(BigInteger(buff.readByteArray()), buff.readDouble())
		}
	}
}

fun FriendlyByteBuf.readImpreciseFraction() = ImpreciseFraction.read(this)
fun FriendlyByteBuf.writeImpreciseFraction(value: ImpreciseFraction) = value.write(this)

fun InputStream.readImpreciseFraction(): ImpreciseFraction {
	val bytes = ByteArray(readVarIntLE())
	read(bytes)
	return ImpreciseFraction(BigInteger(bytes), readDouble())
}

fun OutputStream.writeImpreciseFraction(value: ImpreciseFraction) {
	val bytes = value.whole.toByteArray()
	writeVarIntLE(bytes.size)
	write(bytes)
	writeDouble(value.decimal)
}

fun CompoundTag.getImpreciseFraction(key: String) = ImpreciseFraction.deserializeNBT(this[key])
fun CompoundTag.putImpreciseFraction(key: String, value: ImpreciseFraction) = put(key, value.serializeNBT())

operator fun CompoundTag.set(key: String, value: ImpreciseFraction) = putImpreciseFraction(key, value)

fun Float.toImpreciseFraction() = ImpreciseFraction(this)
fun Double.toImpreciseFraction() = ImpreciseFraction(this)
fun Int.toImpreciseFraction() = ImpreciseFraction(this)
fun Byte.toImpreciseFraction() = ImpreciseFraction(this)
fun Short.toImpreciseFraction() = ImpreciseFraction(this)
fun Long.toImpreciseFraction() = ImpreciseFraction(this)
fun ImpreciseFraction.toImpreciseFraction() = this

class ImpreciseFractionConfigValue(
	parent: ForgeConfigSpec.ConfigValue<String>,
	val minimum: ImpreciseFraction? = null,
	val maximum: ImpreciseFraction? = null,
) : ObservedConfigValue<ImpreciseFraction>(parent) {
	override fun fromString(value: String): ImpreciseFraction? {
		try {
			val parsed = ImpreciseFraction(value)

			if (minimum != null && minimum > parsed) {
				return minimum
			} else if (maximum != null && maximum < parsed) {
				return maximum
			}

			return parsed
		} catch (err: java.lang.NumberFormatException) {
			return null
		}
	}

	override fun toString(value: ImpreciseFraction): Pair<String, ImpreciseFraction> {
		if (minimum != null && minimum > value) {
			return minimum.toString() to minimum
		} else if (maximum != null && maximum < value) {
			return maximum.toString() to maximum
		}

		return value.toString() to value
	}
}

private fun ForgeConfigSpec.Builder.commentRange(minimum: ImpreciseFraction?, maximum: ImpreciseFraction?) {
	if (minimum != null && maximum != null) {
		comment("Range: $minimum ~ $maximum")
	} else if (minimum != null) {
		comment("Range: >= $minimum")
	} else if (maximum != null) {
		comment("Range: <= $maximum")
	}
}

fun ForgeConfigSpec.Builder.defineImpreciseFraction(path: String, defaultValue: ImpreciseFraction, minimum: ImpreciseFraction? = null, maximum: ImpreciseFraction? = null): ImpreciseFractionConfigValue {
	commentRange(minimum, maximum)
	comment("Default: $defaultValue")
	return ImpreciseFractionConfigValue(define(path, defaultValue.toString()), minimum, maximum)
}

fun ForgeConfigSpec.Builder.defineImpreciseFraction(path: List<String>, defaultValue: ImpreciseFraction, minimum: ImpreciseFraction? = null, maximum: ImpreciseFraction? = null): ImpreciseFractionConfigValue {
	commentRange(minimum, maximum)
	comment("Default: $defaultValue")
	return ImpreciseFractionConfigValue(define(path, defaultValue.toString()), minimum, maximum)
}
